"""Tools for loading student-submitted, maybe-invalid code from source files."""

import importlib.util
import os
from os.path import join as pathjoin
from types import ModuleType
from typing import Any, TypeVar

from dill import Unpickler  # type: ignore

from .core import Problem

Output = TypeVar("Output")


class InvalidSubmissionError(BaseException):
    """Something about the submission was invalid."""


class TooManyMatchingSymbols(InvalidSubmissionError):
    """Too many maching symbols were found."""


class NoMatchingSymbol(InvalidSubmissionError, AttributeError):
    """An expected symbol was not found."""


class SubmissionSyntaxError(InvalidSubmissionError, SyntaxError):
    """The submission held an invalid syntax."""


def _load_source_from_path(path: str, name: str = "module") -> Any:
    """Load the python source file found at path, absolute or relative, as a module.

    There's a lot of weird stuff going on in this method with type signatures and
    poorly-documented code that python uses internally for their `import` statement. I
    got this implementation from https://stackoverflow.com/a/67692 and made only small
    modifications to it, but I'm not 100% sure I can explain how it works.
    """
    spec = importlib.util.spec_from_file_location(name, path)

    if spec is None:
        # Based on inspection of the source, I'm not certain how this can happen, but my
        # type checker insists it can. This seems like the most reasonable error to
        # raise.
        raise FileNotFoundError

    mod = importlib.util.module_from_spec(spec)
    try:
        spec.loader.exec_module(mod)  # type: ignore
    except (SyntaxError, NameError) as err:
        raise SubmissionSyntaxError from err  # group all parse errors

    return mod


def _load_attr_from_module(attr: str, module: ModuleType) -> Any:
    """Get a specific symbol from a module."""
    try:
        return module.__getattribute__(attr)
    except AttributeError as err:
        raise NoMatchingSymbol from err


def load_symbol_from_path(path: str, symbol: str) -> Any:
    """Load a specific symbol from a source file found at path, absolute or relative."""
    mod = _load_source_from_path(path)
    return _load_attr_from_module(symbol, mod)


def load_symbol_from_dir(path: str, symbol: str) -> Any:
    """Load a specific symbol from any of the source files in a directory."""
    matching_symbols = []
    for file in os.listdir(path):

        try:
            file_path = pathjoin(path, file)
            matching_symbols.append(load_symbol_from_path(file_path, symbol))
        except (FileNotFoundError, AttributeError, SyntaxError):
            continue

    if len(matching_symbols) > 1:
        raise TooManyMatchingSymbols
    if len(matching_symbols) == 0:
        raise NoMatchingSymbol
    return matching_symbols[0]


class _ProblemUnpickler(Unpickler):  # type: ignore
    """A custom unpickler which will always get the `Problem` class from `aga`.

    This is a hack-ish thing which is required because dill expects us to unpickle an
    object in the some module it was pickled in, so it can then find the object's type
    and use that for instantiation. We want to be able to unpickle the object in any
    type, and we know that we always have a Problem pickled at `problem.pckl`, so we can
    just assert that it's class should be Problem. This is _highly_ unsafe if we are
    unsure of the safety of `problem.pckl`, but pickle/dill is not remotely safe anyway
    with untrusted data.

    This specific solution will break if dill, for some reason, wants to pickle some
    *other* class named "Problem". In that case, I think the best solution will be to
    look into a custom pickler which changes the module name on that end.
    """

    def find_class(self, module: str, name: str) -> Any:
        if name == "Problem":
            return Problem
        return super().find_class(module, name)


def load_problem(root: str, fname: str = "problem.pckl") -> Problem[Output]:
    """Load a problem from the gradescope environment."""
    with open(pathjoin(root, fname), "rb") as problem_pickled:
        out: Problem[Output] = _ProblemUnpickler(problem_pickled).load()
    return out



"""The core entrypoint for the gradescope frontend.

More complete documentation is in `__init__.py`.
"""

from os.path import join as pathjoin
from shutil import copyfileobj
from tempfile import TemporaryDirectory
from typing import Optional
from zipfile import ZipFile

from dill import dump  # type: ignore
from importlib_resources import open_text

from ..core import Output, Problem


class InvalidProblem(BaseException):
    """The Problem failed some golden tests."""


def into_gradescope_zip(problem: Problem[Output], path: Optional[str] = None) -> str:
    """Convert a Problem into a gradescope autograder zip, returning its location.

    This is the high-level entrypoint for this module.
    """
    _check_problem(problem)
    zip_name = _get_zipfile_dest(path, problem)

    with TemporaryDirectory() as tempdir:
        with ZipFile(zip_name, "w") as zip_f:
            path = _copy_resource_to(tempdir, "run_autograder")
            zip_f.write(path, arcname="run_autograder")

            path = _copy_resource_to(tempdir, "setup.sh")
            zip_f.write(path, arcname="setup.sh")

            path = _dump_problem_into_dir(problem, tempdir)
            zip_f.write(path, arcname="problem.pckl")

        return zip_name


def _get_zipfile_dest(path: Optional[str], problem: Problem[Output]) -> str:
    """Determine the destination in which to put the zip file.

    If `path` is none, this is the problem's name; otherwise it's just the provided
    path.
    """
    if path is None:
        return problem.name() + ".zip"
    else:
        return path


def _check_problem(problem: Problem[Output]) -> bool:
    """Check whether `problem` is valid.

    Currently, this just runs the golden tests for problem.
    """
    try:
        problem.run_golden_tests()
        return True
    except AssertionError as err:
        raise InvalidProblem from err


def _copy_resource_to(
    tempdir: str, fname: str, package: str = "aga.gradescope.resources"
) -> str:
    """Copy the resource at package.fname to tempdir/fname, returning the dest path."""
    path = pathjoin(tempdir, fname)
    with open_text(package, fname) as src:
        with open(path, "w", encoding="UTF-8") as dest:
            copyfileobj(src, dest)

    return path


def _dump_problem_into_dir(
    problem: Problem[Output], tempdir: str, fname: str = "problem.pckl"
) -> str:
    """Dump a problem into a directory, returning the pckl file path."""
    path = pathjoin(tempdir, fname)
    _dump_problem_at_path(problem, path)
    return path


def _dump_problem_at_path(problem: Problem[Output], dest: str) -> None:
    """Pickle the problem into a destination."""
    with open(dest, "wb") as file:
        dump(problem, file)



"""The core library functionality.

There are mypy "type: ignore" comments scattered throughout this file. This is because
typing *args and **kwargs is quite difficult. There might be a way to do this, but I'm
not sure how, and it seemed much easier to just ignore them.
"""

from typing import Any, Callable, Dict, Generic, List, Optional, Tuple, TypeVar
from unittest import TestCase, TestSuite

Output = TypeVar("Output")


class _AutograderTestCase(TestCase):
    """A `TestCase` which runs a single test of a `Problem`."""

    def __init__(
        self,
        test_input: "_TestInputs",
        golden: Callable[..., Output],
        under_test: Callable[..., Output],
    ) -> None:
        super().__init__()
        self._test_input = test_input
        self._golden = golden
        self._under_test = under_test

    # pylint: disable=invalid-name
    # camelCase is required by unittest
    def runTest(self) -> None:
        """Run the test case."""
        self._test_input.check(self._golden, self._under_test)

    runTest.__weight__ = 1  # type: ignore

    # pylint: disable=invalid-name
    # camelCase is required by unittest
    def shortDescription(self) -> str:
        """Dynamically generate the test name.

        This method is called by unittest, and then in turn by gradescope_utils, to
        determine what to name the test on gradescope.
        """
        return f"Test {repr(self._test_input)}"


class _TestInputs(TestCase):
    """A single set of test inputs for a problem.

    These will be run against the golden solution and the function under test; their
    outputs will be compared, and a unittest failure raised if they differ.
    """

    def __init__(self, *args, **kwargs) -> None:  # type: ignore
        super().__init__()
        self._args: Tuple = args  # type: ignore
        self._kwargs: Dict = kwargs  # type: ignore

    def _eval(self, func: Callable[..., Output]) -> Output:
        """Evaluate func on the arguments."""
        return func(*self._args, **self._kwargs)

    def check(
        self, golden: Callable[..., Output], under_test: Callable[..., Output]
    ) -> None:
        """Assert that `golden` and `under_test` give the same output.

        This method relies on `TestCase.assertEqual`, which will raise a unittest error
        if they differ. This means it can be plugged into any unittest TestCase as a
        helper method.
        """
        golden_output = self._eval(golden)
        under_test_output = self._eval(under_test)

        self.assertEqual(
            golden_output,
            under_test_output,
            msg=(
                f"Checked with {repr(self)}. Expected {repr(golden_output)}. "
                f"Got {repr(under_test_output)} instead."
            ),
        )

    def generate_test_case(
        self, golden: Callable[..., Output], under_test: Callable[..., Output]
    ) -> _AutograderTestCase:
        """Generate a TestCase which tests `golden` against `under_test`."""
        return _AutograderTestCase(self, golden, under_test)

    def _args_repr(self) -> str:
        return ",".join(repr(x) for x in self._args)

    def _kwargs_repr(self) -> str:
        # we use k instead of repr(k) so we don't get quotes around it
        return ",".join(k + "=" + repr(v) for k, v in self._kwargs.items())

    @staticmethod
    def _repr_sep(args_repr: str, kwargs_repr: str) -> str:
        """Return ',' if both exist, '' otherwise."""
        return args_repr and kwargs_repr and "," or ""

    def __repr__(self) -> str:
        args_repr = self._args_repr()
        kwargs_repr = self._kwargs_repr()
        sep = self._repr_sep(args_repr, kwargs_repr)

        return args_repr + sep + kwargs_repr


class _GoldenTestInputs(_TestInputs, TestCase):
    """A set of test inputs which also contains an expected output.

    This will be treated as a `TestInput` when the autograder is built, but provides
    additional methods to verify a golden solution against the expected output, for
    testing the accuracy of the golden solution against known outputs.
    """

    def __init__(self, output: Output, *args, **kwargs) -> None:  # type: ignore
        super().__init__(*args, **kwargs)
        self._output = output

    def check_one(self, under_test: Callable[..., Output]) -> None:
        """Assert that `under_test`'s output matches the golden output."""
        self.assertEqual(self._eval(under_test), self._output)


class Problem(Generic[Output]):
    """Stores tests for a single problem."""

    def __init__(self, golden: Callable[..., Output], name: str) -> None:
        self._golden: Callable[..., Output] = golden
        self._name = name
        self._test_cases: List[_TestInputs] = []
        self._golden_test_cases: List[_GoldenTestInputs] = []

    def add_test_case(self, case: _TestInputs) -> None:
        """Add a test case to the problem.

        Student solutions will be checked against the golden solution; i.e., this method
        does _not_ produce a test of the golden solution.
        """
        self._test_cases.append(case)

    def add_golden_test_case(self, case: _GoldenTestInputs) -> None:
        """Add a golden test case to the problem.

        Student solutions will be checked against the golden solution, and both against
        the output pasted to the GoldenTestInputs constructor; i.e., this method _does_
        produce a test of the golden solution.
        """
        self._golden_test_cases.append(case)

    def run_golden_tests(self) -> None:
        """Run all tests of the golden solution."""
        for case in self._golden_test_cases:
            case.check_one(self._golden)

    def generate_test_suite(self, under_test: Callable[..., Output]) -> TestSuite:
        """Generate a `TestSuite` for the student submitted function.

        Neither the generated test suite nor the body of this function will run golden
        tests; instead, golden test cases are treated as equivalent to ordinary ones. To
        test the golden function, `run_golden_tests` should be used instead.

        Parameters
        ----------
        under_test : Callable[..., Output]
            The student submitted function.

        Returns
        -------
        TestSuite
            A unittest test suite containing one test for each TestInput in this
            problem, checking the result of the problem's golden solution against
            `under_test`.
        """
        suite = TestSuite([])

        for case in self._test_cases:
            suite.addTest(case.generate_test_case(self._golden, under_test))

        for case in self._golden_test_cases:
            suite.addTest(case.generate_test_case(self._golden, under_test))

        return suite

    def name(self) -> str:
        """Get the problem's name."""
        return self._name

    def expected_symbol(self) -> str:
        """Get the name of the symbol that should be tested against."""
        return self._golden.__name__


def problem(
    name: Optional[str] = None,
) -> Callable[[Callable[..., Output]], Problem[Output]]:
    """Declare a function as the golden solution to a problem.

    This method should decorate a function which is known to produce the correct
    outputs, which we refer to as the "golden solution". It also provides facilities
    for testing that solution, via golden test cases, constructed by passing the
    output argument to the test_case decorator.

    Parameters
    ----------
    name : Optional[str]
        The problem's name. If None (the default), the wrapped function's name will be
        used.

    Returns
    -------
    Callable[[Callable[..., T]], Problem[T]]
        A decorator which turns a golden solution into a problem.
    """

    def outer(func: Callable[..., Output]) -> Problem[Output]:
        problem_name = name or func.__name__

        return Problem(func, problem_name)

    return outer


def _check_reserved_keyword(kwd: str) -> None:
    """Raise an error if `kwd` is reserved."""
    if kwd.startswith("aga_"):
        raise ValueError(
            f"invalid keyword arg to `test_case`: {kwd}: all keyword args"
            "beginning `aga_` are reserved"
        )


def test_case(  # type: ignore
    *args, aga_output: Optional[Any] = None, **kwargs
) -> Callable[[Problem[Output]], Problem[Output]]:
    r"""Declare a specific test case for some problem.

    Parameters
    ----------
    args :
        The arguments to be passed to the functions under test.
    aga_output : Optional[T]
        If aga_output is None, the inputs will be tested against the wrapped function,
        the "golden solution" to the problem. If aga_output is specified, the inputs
        will double as a test _of_ the golden solution; to successfully produce the
        problem grader, the golden solution must return aga_output from the given input.
    kwargs :
        Keyword arguments to be passed to the functions under test. Any keyword starting
        with aga\_ is reserved.

    Returns
    -------
    Callable[[Problem[T]], Problem[T]]
        A decorator which adds the test case to a problem.
    """
    for kwd, _ in kwargs.items():
        _check_reserved_keyword(kwd)

    def outer(prob: Problem[Output]) -> Problem[Output]:
        if aga_output is not None:
            prob.add_golden_test_case(_GoldenTestInputs(aga_output, *args, **kwargs))

        else:
            prob.add_test_case(_TestInputs(*args, **kwargs))

        return prob

    return outer



"""Contains the command-line interface."""

from typing import Iterable, Optional, Tuple

import typer

from .core import Output, Problem
from .gradescope import into_gradescope_zip
from .loader import NoMatchingSymbol, TooManyMatchingSymbols, load_symbol_from_dir

app = typer.Typer()

FRONTENDS = (
    (
        "gradescope",
        "generate a gradescope autograder zip",
    ),
)


def complete_frontend(incomplete: str) -> Iterable[Tuple[str, str]]:
    """Autocomplete a frontend."""
    for frontend in FRONTENDS:
        if frontend[0].startswith(incomplete):
            yield frontend


def _gen_gradescope(problem: Problem[Output], path: Optional[str] = None) -> None:
    """Generate a Gradescope autograder zip."""
    zip_path = into_gradescope_zip(problem, path=path)
    typer.echo(zip_path)


def _load_problem(problem: str) -> Problem[Output]:
    """Load a problem from the top-level directory."""
    try:
        return load_symbol_from_dir(".", problem)  # type: ignore

    except NoMatchingSymbol as err:
        typer.echo(f"problem not found: {problem}", err=True)
        raise typer.Exit(1) from err
    except TooManyMatchingSymbols as err:
        typer.echo(f"multiple matching problems: {problem}", err=True)
        raise typer.Exit(1) from err


# def _process_frontend(frontend: str) -> None:
#     if frontend not in (f[0] for f in FRONTENDS):
#         _handle_invalid_frontend(frontend)


def _handle_invalid_frontend(frontend: str) -> None:
    """Error on an invalid frontend."""
    typer.echo(f"invalid frontend: {frontend}", err=True)
    raise typer.Exit(1)


@app.command()
def gen(
    problem_name: str = typer.Argument(
        ..., help='The problem to generate (see "problem discovery" in the CLI docs).'
    ),
    frontend: str = typer.Option(
        "gradescope",
        autocompletion=complete_frontend,
        help="The frontend to use. Currently only gradescope is supported.",
    ),
    dest: Optional[str] = typer.Option(
        None, help="The path to place the output file(s)."
    ),
) -> None:
    """Generate an autograder file for a problem."""
    problem = _load_problem(problem_name)  # type: ignore

    if frontend == "gradescope":
        _gen_gradescope(problem, dest)
    else:
        _handle_invalid_frontend(frontend)


@app.command()
def check(
    problem_name: str = typer.Argument(
        ..., help='The problem to check (see "problem discovery" in the CLI docs).'
    ),
) -> None:
    """Check a problem against test cases with an `aga_output`."""
    problem = _load_problem(problem_name)  # type: ignore

    try:
        problem.run_golden_tests()
    except AssertionError as err:
        typer.echo(f"{problem.name()} failed some golden tests: {err}", err=True)
        raise typer.Exit(1) from err
    else:
        typer.echo(f"{problem.name()} passed golden tests.")
        raise typer.Exit()


click_object = typer.main.get_command(app)  # exposed for documentation



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['aga', 'aga.gradescope', 'aga.gradescope.resources']

package_data = \
{'': ['*']}

install_requires = \
['dill>=0.3.4,<0.4.0',
 'gradescope-utils>=0.4.0,<0.5.0',
 'importlib-metadata>=4.8.1,<5.0.0',
 'importlib-resources>=5.2.2,<6.0.0',
 'typer>=0.4.0,<0.5.0']

entry_points = \
{'console_scripts': ['aga = aga.cli:app']}

setup_kwargs = {
    'name': 'aga',
    'version': '0.4.0',
    'description': 'aga grades assignments',
    'long_description': '# aga\n\n[![Tests](https://github.com/nihilistkitten/aga/workflows/Tests/badge.svg)](https://github.com/nihilistkitten/aga/actions?workflow=Tests)\n[![Codecov](https://codecov.io/gh/nihilistkitten/aga/branch/main/graph/badge.svg)](https://codecov.io/gh/nihilistkitten/aga)\n[![PyPI](https://img.shields.io/pypi/v/aga.svg)](https://pypi.org/project/aga/)\n[![Read the Docs](https://readthedocs.org/projects/aga/badge/)](https://aga.readthedocs.io/)\n[![License](https://img.shields.io/github/license/nihilistkitten/aga)](https://choosealicense.com/licenses/mit/)\n\n**aga** (aga grades assignments) is a tool for easily producing autograders for python programming assignments.\n\n## Installation\n\nInstall from pip:\n\n```bash\npip install aga\n```\n\n## Quickstart\n\nIn `square.py` (or any python file), write:\n\n```python\nfrom aga import problem, test_case\n\n\n@test_case(-3)\n@test_case(100)\n@test_case(2, aga_output=4)\n@test_case(-2, aga_output=4)\n@problem()\ndef square(x: int) -> int:\n    """Square x."""\n    return x * x\n```\n\nThen run `aga gen square` from the directory with `square.py`. This will generate a ZIP file suitable for upload to gradescope.\n\n## Usage\n\nAga relies on the notion of a _golden solution_ to a given problem which is known to be correct. The main work of the library is to compare the output of this golden solution on some family of test inputs against the output of a student submission. To that end, aga integrates with frontends: existing classroom software which allow submission of student code. Currently, only gradescope is supported.\n\nTo use aga:\n\n1. Write a golden solution to some programming problem.\n2. Decorate this solution with the `problem` decorator.\n3. Decorate this problem with any number of `test_case` decorators, which take arbitrary positional or keyword arguments and pass them verbatim to the golden and submitted functions.\n4. Generate the autograder using the CLI: `aga gen <function_name>`.\n\nThe `test_case` decorator may optionally take a special keyword argument called `aga_output`. This allows easy testing of the golden solution: aga will not successfully produce an autograder unless the golden solution\'s output matches the `aga_output`. You should use these as sanity checks to ensure your golden solution is implemented correctly.\n\nFor complete documentation, including configuring problem and test case metadata, see the [API reference](https://aga.readthedocs.io/en/stable/reference.html).\n\nFor CLI documentation, run `aga --help`, or access the docs [online](https://aga.readthedocs.io/en/stable/cli.html).\n\n## Contributing\n\nBug reports, feature requests, and pull requests are all welcome. For details on our test suite, development environment, and more, see the [developer documentation](https://aga.readthedocs.io/en/stable/development.html).\n',
    'author': 'Riley Shahar',
    'author_email': 'riley.shahar@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/nihilistkitten/aga',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6.1,<4.0.0',
}


setup(**setup_kwargs)
